import time
import logging
import logging.handlers
import socket
import threading
import yaml
from urllib.parse import parse_qs
import queue  # Import the queue module for thread-safe queues

# Initialize global variables and locks
lock = threading.Lock()
shutdown_event = threading.Event()
queries_info = {}
active_threads = 0
zmapi = None
monitors = None

alerted = []
alerted_lock = threading.Lock()

network_lock = threading.Lock()

# Initialize two thread-safe queues
queue1 = queue.Queue()
queue2 = queue.Queue()

import os

# Define the path to the config file
config_file_path = '/etc/zm/trigserv.yaml'

# Check if the config file exists
if not os.path.exists(config_file_path):
    print(f"Error: The configuration file {config_file_path} was not found.")
    exit(1)  # Exit if the config file does not exist

# Load configuration from YAML file
with open(config_file_path, 'r') as config_file:
    config = yaml.safe_load(config_file)

if config is None:
    print("Error: Configuration is empty or incorrectly formatted.")
    exit(1)  # Exit if the configuration is not loaded properly

def setup_logging():
    if 'logpath' in config:
        log_filename = config['logpath']
    else:
        print("Logpath not found in config. Using default 'app.log'.")
        log_filename = 'app.log'
    
    log_file_size_limit = config.get('log_file_size_limit', 1048576)  # Use a default value if not specified
    
    # Ensure the log directory exists
    os.makedirs(os.path.dirname(log_filename), exist_ok=True)
    
    # Set up a logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    
    # Create handlers
    c_handler = logging.StreamHandler()
    f_handler = logging.handlers.RotatingFileHandler(
        log_filename, maxBytes=log_file_size_limit, backupCount=5)
    
    # Create formatters and add it to handlers
    c_format = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    f_format = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    c_handler.setFormatter(c_format)
    f_handler.setFormatter(f_format)
    
    # Add handlers to the logger
    logger.addHandler(c_handler)
    logger.addHandler(f_handler)
    
    return logger


logger = setup_logging()

def send_command(command):
    try:
        with network_lock:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                # Connect to the server with a timeout
                s.settimeout(5)  # 5 seconds timeout
                s.connect((config.get("zmtriggerhost", "localhost"), config.get("zmtriggerport", 6802)))
                # Send the command followed by a newline, and encode to bytes
                s.sendall(f"{command}\n".encode())
    except socket.timeout:
        logger.error(f"Timeout occurred when sending command: {command}")
    except Exception as e:
        logger.error(f"An error occurred when sending command: {command}. Error: {str(e)}")

def alert_on(monitor, cause, score=1, info='',showtext=''):
    if monitor in config.get("monitors",{}).keys():
        id = config.get("monitors",{})[monitor]['id']
        command = str(id)+"|on|"+str(score)+"|"+cause+"|"+info+"|"+showtext
        send_command(command)
    else:
        logger.error("Attempted to switch on a monitor that doesn't exist! Monitor name:"+monitor)

def alert_off(monitor, cause, score=1, info='',showtext=''):
    if monitor in config.get("monitors",{}).keys():
        id = config.get("monitors",{})[monitor]['id']
        command = str(id)+"|cancel||||"
        send_command(command)
    else:
        logger.error("Attempted to switch off a monitor that doesn't exist! Monitor name:"+monitor)

def handle_client_connection(client_socket, client_address):
    global active_threads
    try:
        request = client_socket.recv(1024).decode('utf-8')
        logger.info(f"Received request: {request}")

        # Extract the path and query string from the request
        path = request.split(' ')[1]
        query_string = path.split('?')[1] if '?' in path else ""
        query_vars = parse_qs(query_string)

        # Retrieve username and userpwd from the query variables
        username = query_vars.get('username', [None])[0]
        userpwd = query_vars.get('userpwd', [None])[0]

        # Check if the credentials match those in the config file
        if username == config.get('user') and userpwd == config.get('password'):
            logger.info(f"User {username} authenticated successfully.")
            # Proceed with the function's main tasks here
            # For example, handling client's request based on the authenticated user
        else:
            logger.warning(f"Authentication failed for user {username}. from ip:{client_address[0]}")
            return  # Return immediately if authentication fails

        query_data = {
            "query_vars": query_vars,
            "timestamp": int(time.time())
        }

        with lock:
            queries_info[client_address[0]] = query_data
    finally:
        with lock:
            active_threads -= 1
        client_socket.close()

    monitor = ''
    for mon in config.get('monitors', dict()).keys():
        if config.get('monitors', dict())[mon]['ip'] == client_address[0]:
            monitor = mon
    with alerted_lock:
        if monitor and monitor not in alerted:
            queue1.put(monitor)


def check_ongoing():
    while not shutdown_event.is_set():
        with alerted_lock:
            with lock:
                for alerted_item in alerted:
                    if alerted_item in config.get('monitors', dict()).keys() and \
                        config.get('monitors', dict())[alerted_item]['ip'] in queries_info.keys() and \
                        int(time.time()) - queries_info[config.get('monitors', dict())[alerted_item]['ip']]['timestamp'] > config.get('monitors', dict())[alerted_item]['expiry']:
                        queue2.put(alerted_item)
        shutdown_event.wait(10)  # Wait for 10 seconds or until the shutdown event is set

def notifier():
    while not shutdown_event.is_set():

        try:
            # Process items from queue2
            item2 = queue2.get_nowait()
            logger.info(f"Cancelling alert on: {item2}")
            alert_off(item1, "Cam Event")
            queue2.task_done()
            with alerted_lock:
                alerted.remove(item2)
        except queue.Empty:
            pass
        except Exception:
            logger.error("Fail in cancel alert queue processing!")

        try:
            # Process items from queue1
            item1 = queue1.get_nowait()
            logger.info(f"Alerting on: {item1}")
            alert_on(item1, "Cam Event")
            queue1.task_done()
            with alerted_lock:
                alerted.append(item1)
        except queue.Empty:
            pass
        except Exception:
            logger.error("Fail in alert queue processing!")
        # Wait for an item to be added to one of the queues
        if queue1.empty() and queue2.empty():
            time.sleep(0.1)  # Polling interval

def start_server():
    global active_threads
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server.bind(('0.0.0.0', 8080))
    server.listen(5)
    logger.info("Listening on port 8080...")

    alive_thread = threading.Thread(target=check_ongoing)
    alive_thread.start()

    notifier_thread = threading.Thread(target=notifier)
    notifier_thread.start()

    try:
        while not shutdown_event.is_set():
            server.settimeout(1.0)  # Timeout for the accept call
            try:
                client_sock, client_address = server.accept()
            except socket.timeout:
                continue  # Go back to the top of the loop

            with lock:
                if active_threads < config.get('max_threads', 20):
                    active_threads += 1
                    client_handler = threading.Thread(target=handle_client_connection, args=(client_sock, client_address))
                    client_handler.start()
                else:
                    logger.warninging("Max threads reached, waiting before accepting new connections.")
    except KeyboardInterrupt:
        logger.info("Shutting down...")
        shutdown_event.set()
        alive_thread.join()
        notifier_thread.join()

    server.close()

if __name__ == "__main__":
    start_server()
